#include <engine/allocmem.h>
#include <engine/asset.h>
#include <engine/display.h>
#include <engine/2d/transform.h>
#include <engine/3d/render.h>

/*
 * mol_alloc:
 *	Reads a MOL file from a buffer in system memory onto
 *	a given AssetMOL structure. MOL files hold 2D collision
 *	and map data.
 */
void
mol_alloc(p_mol, p_data)
struct AssetMOL *p_mol;
void *p_data;
{
	int i, j;
	unsigned char	*p_ptr = p_data;
	unsigned int	bufint = 0;

	/* HEADER section */
	p_mol->HEADER.id	= *(p_ptr);
	p_mol->HEADER.version	= *(p_ptr+1);
	p_mol->HEADER.tile_w	= *(p_ptr+2);
	p_mol->HEADER.tile_h	= *(p_ptr+3);

	p_ptr += 4;

	/* NOTE: supports latest version only */
	if(p_mol->HEADER.version == MOL_LATEST_VERSION)
	{
		/* COLMAP section */
		p_mol->COLMAP.bnum = *((unsigned int *)(p_ptr));

		p_ptr += 4;

		/* allocate memory for colliders */
		{
			unsigned int total_colliders = (p_mol->COLMAP.bnum-4)/MOL_COLMAP_ELEM_SIZE;

			p_mol->COLMAP.p_elem = (struct COLMAP_Collider *)allocmem(ALLOCBUF, total_colliders, MOL_COLMAP_ELEM_SIZE);

			if(!p_mol->COLMAP.p_elem)
			{
				printf("mol_alloc() error: could not allocate memory for p_mol->COLMAP.p_elem\n");
				while(1);
			}

			/* read COLMAP element data */
			for(i = 0; i < total_colliders; i++)
			{
				for(j = 0; j < 5; j++)
				{
					bufint = *((unsigned int *)(p_ptr));
					switch(j)
					{
						case 0:
							p_mol->COLMAP.p_elem[i].shape = bufint;
							break;
						case 1:
							p_mol->COLMAP.p_elem[i].x = bufint;
							break;
						case 2:
							p_mol->COLMAP.p_elem[i].y = bufint;
							break;
						case 3:
							p_mol->COLMAP.p_elem[i].w = bufint;
							break;
						case 4:
							p_mol->COLMAP.p_elem[i].h = bufint;
							break;
					}
					p_ptr += 4;
				}
			}
		}

		/* AREATRIG section */
		bufint = *((unsigned int *)(p_ptr));

		{
			unsigned int total_triggers = (bufint-4)/MOL_AREATRIG_ELEM_SIZE;

			p_mol->AREATRIG.bnum = bufint;
			p_mol->AREATRIG.p_elem = (struct AREATRIG_Trigger *)allocmem(ALLOCBUF, total_triggers, MOL_AREATRIG_ELEM_SIZE);

			if(!p_mol->AREATRIG.p_elem)
			{
				printf("mol_alloc() error: could not allocate memory for p_mol->AREATRIG.p_elem\n");
				while(1);
			}

			p_ptr += 4;

			for(i = 0; i < total_triggers; i++)
			{
				for(j = 0; j < 4; j++)
				{
					bufint = *((unsigned int *)(p_ptr));
					switch(j)
					{
						case 0:
							p_mol->AREATRIG.p_elem[i].x = bufint;
							break;
						case 1:
							p_mol->AREATRIG.p_elem[i].y = bufint;
							break;
						case 2:
							p_mol->AREATRIG.p_elem[i].w = bufint;
							break;
						case 3:
							p_mol->AREATRIG.p_elem[i].h = bufint;
							break;
					}


					p_ptr += 4;
				}
			}
		}

		/* read LYRHEAD and its LAYER subsection(s) */
		p_mol->LYRHEAD.numlayers = *(p_ptr);
		p_ptr += 4;

		for(i = 0; i < p_mol->LYRHEAD.numlayers; i++)
		{
			/* set bnum */
			p_mol->LYRHEAD.layer[i].bnum = *((unsigned int *)(p_ptr));

			p_ptr += 4;

			p_mol->LYRHEAD.layer[i].width		= *(p_ptr);
			p_mol->LYRHEAD.layer[i].height		= *(p_ptr+1);
			p_mol->LYRHEAD.layer[i].is_sparse	= *(p_ptr+2);

			p_ptr += 4;

			/* allocate & read Coordinated List */
			if(p_mol->LYRHEAD.layer[i].is_sparse)
			{
				int total_elements = (p_mol->LYRHEAD.layer[i].bnum - 8)/3;

				p_mol->LYRHEAD.layer[i].p_coo = (struct LayerCoordinatedList *)allocmem(ALLOCBUF, total_elements, 3);

				if(!p_mol->LYRHEAD.layer[i].p_coo)
				{
					printf("mol_alloc() error: could not allocate memory for p_mol->LYRHEAD.layer[i].p_coo\n");
					while(1);
				}
				for(j = 0; j < total_elements; j++)
				{
					p_mol->LYRHEAD.layer[i].p_coo[j].val = *(p_ptr);
					p_mol->LYRHEAD.layer[i].p_coo[j].row = *(p_ptr+1);
					p_mol->LYRHEAD.layer[i].p_coo[j].col = *(p_ptr+2);

					p_ptr += 3;
				}
			}
			/* allocate & read normal matrix */
			else
			{
				int total_elements = p_mol->LYRHEAD.layer[i].bnum - 8;
				int max_rows = p_mol->LYRHEAD.layer[i].height;
				int max_cols = p_mol->LYRHEAD.layer[i].width;
				int row = 0, col = 0;

				p_mol->LYRHEAD.layer[i].p_matrix = (struct LayerMatrix *)allocmem(ALLOCBUF, total_elements, 1);

				if(!p_mol->LYRHEAD.layer[i].p_matrix)
				{
					printf("mol_alloc() error: could not allocate memory for p_mol->LYRHEAD.layer[i].p_matrix\n");
					while(1);
				}

				for(row = 0; row < max_rows; row++)
				{
					for(col = 0; col < max_cols; col++)
					{
						p_mol->LYRHEAD.layer[i].p_matrix[row*max_cols + col].val = *(p_ptr);

						p_ptr++;
					}
				}
			}
		}
	}
	else
	{
		printf("mol_alloc() error: Unsupported version of mol! (v%d)\n", p_mol->HEADER.version);
		while(1);
	}
}

/*
 * mol_free:
 *	Frees the memory allocated by mol_alloc.
 */
void
mol_free(p_mol)
struct AssetMOL *p_mol;
{
	int i;

	allocmem_free(ALLOCBUF, p_mol->COLMAP.p_elem);
	allocmem_free(ALLOCBUF, p_mol->AREATRIG.p_elem);
	allocmem_free(ALLOCBUF, p_mol->AREATRIG.p_elem);

	for(i = 0; i < p_mol->LYRHEAD.numlayers; i++)
	{
		if(p_mol->LYRHEAD.layer[i].is_sparse)
		{
			allocmem_free(ALLOCBUF, p_mol->LYRHEAD.layer[i].p_coo);
		}
		else
		{
			allocmem_free(ALLOCBUF, p_mol->LYRHEAD.layer[i].p_matrix);
		}
	}
}

#define BLOCK_SIZE_POLY		12 /* POLY_F3, POLY_F4, POLY_G3, POLY_G4 */
#define BLOCK_SIZE_POLYT	20 /* POLY_FT3, POLY_FT4, POLY_GT3, POLY_GT4 */
/*
 * determine_block_size:
 *	Helper function for determining the size of a single polygon block
 *	in the POLYGON section.
 */
static unsigned int
determine_block_size(flag)
unsigned short flag;
{
	int block_size = BLOCK_SIZE_POLY;

	if(flag & XASS_FLAG_TEXTURED)
	{
		block_size+=8;
	}

	/* flat-shaded lighting */
	if(!(flag & XASS_FLAG_GOURAUD))
	{
		if(flag & XASS_FLAG_LIGHTING)
		{
			block_size+=8;
		}
	}
	else
	{
		block_size += 8; /* for RGB section, since all gouraud polygons require it */

		if(flag & XASS_FLAG_LIGHTING)
		{
			block_size+=8;
		}
	}

	return block_size;
}

/*
 * mtd_alloc:
 *	Reads a MTD file from a buffer in system memory onto
 *	a given AssetMTD structure. MTD files hold 3D model
 *	data without animation.
 */
void
mtd_alloc(p_mtd, p_data)
struct AssetMTD *p_mtd;
void *p_data;
{
#define BLOCK_SIZE_LINEF2	8

	int			i, j;
	unsigned char		*p_ptr = (unsigned char *)p_data;
	CVECTOR			*p_rgb;
	unsigned int		rgb_section_bnum;

	unsigned short		polyf3_index = 0;
	unsigned short		polyf4_index = 0;
	unsigned short		polyft3_index = 0;
	unsigned short		polyft4_index = 0;

	unsigned short		polyg3_index = 0;
	unsigned short		polyg4_index = 0;
	unsigned short		polygt3_index = 0;
	unsigned short		polygt4_index = 0;
	struct Vector2u		uv[USHRT_MAX];

	setVector(&p_mtd->rotmod2, 0, 0, 0);

	/* init */
	p_mtd->linef2_cnt = 0;
	p_mtd->polyf3_cnt = 0;
	p_mtd->polyf4_cnt = 0;
	p_mtd->polyft3_cnt = 0;
	p_mtd->polyft4_cnt = 0;
	p_mtd->polyg3_cnt = 0;
	p_mtd->polyg4_cnt = 0;
	p_mtd->polygt3_cnt = 0;
	p_mtd->polygt4_cnt = 0;

	/* set header information */
	p_mtd->id = *((unsigned int *)(p_ptr));
	p_ptr += 4;

	p_mtd->version = *((unsigned short *)(p_ptr));
	p_ptr += 2;

	/* NOTE: supports latest version only */
	if(p_mtd->version == MTD_LATEST_VERSION)
	{
		p_mtd->flag = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		p_mtd->nvert = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		if(p_mtd->nvert > MAX_VERTS)
		{
			printf("mtd_alloc() error: number of vertices exceed maximum amount (%d/%d)\n", p_mtd->nvert, MAX_VERTS);
			while(1);
		}

		p_mtd->npoly = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		if(p_mtd->flag & XASS_FLAG_WIREFRAME)
		{
			printf("wireframe:1\n");
			printf("npoly:%d\n", p_mtd->npoly);
		}

		/* init and set vertex information */
		p_mtd->p_vertex = (VECTOR *)allocmem(ALLOCBUF, p_mtd->nvert, sizeof(VECTOR));
		if(!p_mtd->p_vertex)
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->p_vertex\n");
			printf("bytes requested: %d\n", p_mtd->nvert*sizeof(VECTOR));
			while(1);
		}

		for(i = 0; i < p_mtd->nvert; i++, p_ptr += 12)
		{
			p_mtd->p_vertex[i].vx = *((unsigned int *)(p_ptr));
			p_mtd->p_vertex[i].vy = *((unsigned int *)(p_ptr+4));
			p_mtd->p_vertex[i].vz = *((unsigned int *)(p_ptr+8));
		}

		/*
		 * store the UV section for later when we set the UVs
		 * of our primitives
		 */

		if(!(p_mtd->flag & XASS_FLAG_WIREFRAME))
		{
			if(p_mtd->flag & XASS_FLAG_TEXTURED)
			{
				unsigned int bnum = *((unsigned int *)(p_ptr));
				p_ptr += 4;

				/* FIXME: instead of storing everything into this local uv buffer, just use a fucking pointer */
				for(i = 0; i < (bnum-4)/2; i++, p_ptr += 2)
				{
					uv[i].u = *(p_ptr);
					uv[i].v = *(p_ptr+1);
				}

				if((i*2)%4) /* the iterator's value after a for loop in C89 and Python 3.4.2 (via range()) are not the same :p */
				{
					p_ptr += 2;
				}
			}

			/*
			 * NORMAL section must be allocated in a separate memory location
			 * than our primitives since they do not store normals.
			 */
			if(p_mtd->flag & XASS_FLAG_LIGHTING)
			{
				unsigned int bnum = *((unsigned int *)(p_ptr));
				unsigned int nnorm = (bnum-4)/12;

				printf("nnorm: %d; bnum: %d\n", nnorm, bnum);

				p_ptr += 4;

				p_mtd->p_normal = (VECTOR *)allocmem(ALLOCBUF, nnorm, sizeof(VECTOR));
				if(!p_mtd->p_normal)
				{
					printf("mtd_alloc() error: could not allocate memory for p_mtd->p_normal\n");
					while(1);
				}

				for(i = 0; i < nnorm; i++, p_ptr += 12)
				{
					p_mtd->p_normal[i].vx = *((int *)(p_ptr));
					p_mtd->p_normal[i].vy = *((int *)(p_ptr+4));
					p_mtd->p_normal[i].vz = *((int *)(p_ptr+8));
				}
			}

			/* RGB section */
			if(p_mtd->flag & XASS_FLAG_GOURAUD)
			{
				rgb_section_bnum = *((unsigned int *)(p_ptr));
				p_rgb = p_ptr + 4;

				p_ptr += rgb_section_bnum;
			}
		}

		/* initialize p_indices array */
		allocmem_align(ALLOCBUF);
		p_mtd->p_indices = (struct AVInd *)allocmem(ALLOCBUF, p_mtd->npoly, sizeof(struct AVInd));

		if(!p_mtd->p_indices)
		{
			allocmem_print(ALLOCBUF);
			printf("mtd_alloc() error: could not allocate memory for p_mtd->p_indices (0x%.8x)\n", p_mtd->p_indices);
			printf("bytes requested: %d\nallocmem_error code %d\n", p_mtd->npoly*sizeof(struct AVInd), allocmem_error);
			while(1);
		}

		/*
		 * set polygon information
		 * first, count how many primitives of each type there are
		 */
		for(i = 0; i < p_mtd->npoly; i++)
		{
			unsigned char	type = 0;
			unsigned int	block_size = determine_block_size(p_mtd->flag);

			if(p_mtd->flag & XASS_FLAG_WIREFRAME) /* use LINE_F2s, period. no mix of tris and quads */
			{
				p_mtd->linef2_cnt = p_mtd->npoly;
				break;
			}
			else /* normal quads and tris */
			{
				type = *( p_ptr + i*block_size );

				switch(type)
				{
					case 0: /* tri */
						if(p_mtd->flag & XASS_FLAG_GOURAUD)
						{
							if(p_mtd->flag & XASS_FLAG_TEXTURED)
							{
								p_mtd->polygt3_cnt++;
							}
							else
							{
								p_mtd->polyg3_cnt++;
							}
						}
						else
						{
							if(p_mtd->flag & XASS_FLAG_TEXTURED)
							{
								p_mtd->polyft3_cnt++;
							}
							else
							{
								p_mtd->polyf3_cnt++;
							}
						}
						p_mtd->p_indices[i].type = 0;

						break;
					case 1: /* quad */
						if(p_mtd->flag & XASS_FLAG_GOURAUD)
						{
							if(p_mtd->flag & XASS_FLAG_TEXTURED)
							{
								p_mtd->polygt4_cnt++;
							}
							else
							{
								p_mtd->polyg4_cnt++;
							}
						}
						else
						{
							if(p_mtd->flag & XASS_FLAG_TEXTURED)
							{
								p_mtd->polyft4_cnt++;
							}
							else
							{
								p_mtd->polyf4_cnt++;
							}
						}
						p_mtd->p_indices[i].type = 1;

						break;
					default:
						printf("mtd_alloc() error: unknown polygon type! (%d; on iter %d @ offset 0x%x)\n", type, i, p_ptr-(unsigned char *)p_data);
						while(1);
				}
			}
		}

		/* allocate polygon primitive buffers */
		p_mtd->apb.p_linef2 = (LINE_F2 *)allocmem(ALLOCBUF, p_mtd->linef2_cnt, sizeof(LINE_F2));
		p_mtd->apb.p_polyf3 = (POLY_F3 *)allocmem(ALLOCBUF, p_mtd->polyf3_cnt, sizeof(POLY_F3));
		p_mtd->apb.p_polyf4 = (POLY_F4 *)allocmem(ALLOCBUF, p_mtd->polyf4_cnt, sizeof(POLY_F4));
		p_mtd->apb.p_polyft3 = (POLY_FT3 *)allocmem(ALLOCBUF, p_mtd->polyft3_cnt, sizeof(POLY_FT3));
		p_mtd->apb.p_polyft4 = (POLY_FT4 *)allocmem(ALLOCBUF, p_mtd->polyft4_cnt, sizeof(POLY_FT4));

		p_mtd->apb.p_polyg3 = (POLY_G3 *)allocmem(ALLOCBUF, p_mtd->polyg3_cnt, sizeof(POLY_G3));
		p_mtd->apb.p_polyg4 = (POLY_G4 *)allocmem(ALLOCBUF, p_mtd->polyg4_cnt, sizeof(POLY_G4));
		p_mtd->apb.p_polygt3 = (POLY_GT3 *)allocmem(ALLOCBUF, p_mtd->polygt3_cnt, sizeof(POLY_GT3));
		p_mtd->apb.p_polygt4 = (POLY_GT4 *)allocmem(ALLOCBUF, p_mtd->polygt4_cnt, sizeof(POLY_GT4));

		/* flat */
		if( (p_mtd->linef2_cnt > 0) && (!p_mtd->apb.p_linef2) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_linef2\n");
			while(1);
		}

		if( (p_mtd->polyf3_cnt > 0) && (!p_mtd->apb.p_polyf3) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyf3\n");
			while(1);
		}

		if( (p_mtd->polyf4_cnt > 0) && !(p_mtd->apb.p_polyf4) )
		{
			printf("bytes requested: %d\n", p_mtd->polyf4_cnt*sizeof(POLY_F4));
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyf4\n");
			while(1);
		}

		if( (p_mtd->polyft3_cnt > 0) && (!p_mtd->apb.p_polyft3) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyft3\n");
			while(1);
		}

		if( (p_mtd->polyft4_cnt > 0) && !(p_mtd->apb.p_polyft4) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyft4\n");
			while(1);
		}

		/* gouraud */
		if( (p_mtd->polyg3_cnt > 0) && (!p_mtd->apb.p_polyg3) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyg3\n");
			while(1);
		}

		if( (p_mtd->polyg4_cnt > 0) && !(p_mtd->apb.p_polyg4) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polyg4\n");
			while(1);
		}

		if( (p_mtd->polygt3_cnt > 0) && (!p_mtd->apb.p_polygt3) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polygt3\n");
			while(1);
		}

		if( (p_mtd->polygt4_cnt > 0) && !(p_mtd->apb.p_polygt4) )
		{
			printf("mtd_alloc() error: could not allocate memory for p_mtd->apb.p_polygt4\n");
			while(1);
		}

		/* initialize the primitive buffers */
		for(i = 0; i < p_mtd->linef2_cnt; i++)
		{
			setLineF2(p_mtd->apb.p_linef2 + i);
		}

		for(i = 0; i < p_mtd->polyf3_cnt; i++)
		{
			setPolyF3(p_mtd->apb.p_polyf3 + i);
		}

		for(i = 0; i < p_mtd->polyf4_cnt; i++)
		{
			setPolyF4(p_mtd->apb.p_polyf4 + i);
		}

		for(i = 0; i < p_mtd->polyft3_cnt; i++)
		{
			setPolyFT3(p_mtd->apb.p_polyft3 + i);
		}

		for(i = 0; i < p_mtd->polyft4_cnt; i++)
		{
			setPolyFT4(p_mtd->apb.p_polyft4 + i);
		}

		for(i = 0; i < p_mtd->polyg3_cnt; i++)
		{
			setPolyG3(p_mtd->apb.p_polyg3 + i);
		}

		for(i = 0; i < p_mtd->polyg4_cnt; i++)
		{
			setPolyG4(p_mtd->apb.p_polyg4 + i);
		}

		for(i = 0; i < p_mtd->polygt3_cnt; i++)
		{
			setPolyGT3(p_mtd->apb.p_polygt3 + i);
		}

		for(i = 0; i < p_mtd->polygt4_cnt; i++)
		{
			setPolyGT4(p_mtd->apb.p_polygt4 + i);
		}


		/*
		 * Now transfer data of each polygon into handler.
		 * note how we don't transfer coordinate data, as the
		 * coordinates stored in the p_mtd->p_vertex array are
		 * in raw 3D coordinates. the coordinates for a POLY_XX
		 * are in screen coordinates. we only transfer RGB, UV,
		 * and index data.
		 */
		{
			int block_size;
			if(p_mtd->flag & XASS_FLAG_WIREFRAME)
			{
				for(i = 0; i < p_mtd->npoly; i++, p_ptr += BLOCK_SIZE_LINEF2)
				{
					unsigned char	r, g, b;
					unsigned short	vi0, vi1;

					r = *(p_ptr+1);
					g = *(p_ptr+2);
					b = *(p_ptr+3);

					vi0 = *((unsigned short *)(p_ptr+4));
					vi1 = *((unsigned short *)(p_ptr+6));

					//setRGB0(&p_mtd->apb.p_linef2[i], r, g, b);
					setRGB0(&p_mtd->apb.p_linef2[i], 255, 255, 255);

					p_mtd->p_indices[i].v[0] = vi0;
					p_mtd->p_indices[i].v[1] = vi1;
				}
			}
			else
			{
				/*
				 * Polygon sizes differ depending on certain flag
				 * settings.
				 */
				block_size = determine_block_size(p_mtd->flag);

				for(i = 0; i < p_mtd->npoly; i++, p_ptr += block_size)
				{
					/* FIXME: use an array :p */
					unsigned short	vi0, vi1, vi2, vi3;	/* Vertex indices */
					unsigned short	ui0, ui1, ui2, ui3;	/* UV indices */
					unsigned short	ni0, ni1, ni2, ni3;	/* normal indices */
					unsigned short	ri0, ri1, ri2, ri3;	/* rgb indices */
					unsigned char	r, g, b;

					unsigned char	*p_optsect = p_ptr+12;

					r = *(p_ptr+1);
					g = *(p_ptr+2);
					b = *(p_ptr+3);

					vi0 = *((unsigned short *)(p_ptr+4));
					vi1 = *((unsigned short *)(p_ptr+6));
					vi2 = *((unsigned short *)(p_ptr+8));
					vi3 = *((unsigned short *)(p_ptr+10));

					if(p_mtd->flag & XASS_FLAG_TEXTURED)
					{
						ui0 = *((unsigned short *)(p_optsect));
						ui1 = *((unsigned short *)(p_optsect+2));
						ui2 = *((unsigned short *)(p_optsect+4));
						ui3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					if(p_mtd->flag & XASS_FLAG_LIGHTING)
					{
						ni0 = *((unsigned short *)(p_optsect));
						ni1 = *((unsigned short *)(p_optsect+2));
						ni2 = *((unsigned short *)(p_optsect+4));
						ni3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					if(p_mtd->flag & XASS_FLAG_GOURAUD)
					{
						ri0 = *((unsigned short *)(p_optsect));
						ri1 = *((unsigned short *)(p_optsect+2));
						ri2 = *((unsigned short *)(p_optsect+4));
						ri3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					/* set the colour depending on type */
					switch( *p_ptr )
					{
						case 0: /* tri */
							if(p_mtd->flag & XASS_FLAG_GOURAUD)
							{
								if(p_mtd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_mtd->apb.p_polygt3[polygt3_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									setRGB1(&p_mtd->apb.p_polygt3[polygt3_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									setRGB2(&p_mtd->apb.p_polygt3[polygt3_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);

									setUV3(&p_mtd->apb.p_polygt3[polygt3_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui2].u, uv[ui2].v);
									p_mtd->apb.p_polygt3[polygt3_index].tpage	= p_mtd->tpage;
									p_mtd->apb.p_polygt3[polygt3_index].clut	= p_mtd->clut;
									polygt3_index++;
								}
								else
								{
									setRGB0(&p_mtd->apb.p_polyg3[polyg3_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									setRGB1(&p_mtd->apb.p_polyg3[polyg3_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									setRGB2(&p_mtd->apb.p_polyg3[polyg3_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									polyg3_index++;
								}
							}
							else
							{
								if(p_mtd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_mtd->apb.p_polyft3[polyft3_index], r, g, b);
									setUV3(&p_mtd->apb.p_polyft3[polyft3_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui2].u, uv[ui2].v);
									p_mtd->apb.p_polyft3[polyft3_index].tpage	= p_mtd->tpage;
									p_mtd->apb.p_polyft3[polyft3_index].clut	= p_mtd->clut;
									polyft3_index++;
								}
								else
								{
									setRGB0(&p_mtd->apb.p_polyf3[polyf3_index], r, g, b);
									polyf3_index++;
								}
							}
							break;
						case 1: /* quad */
							if(p_mtd->flag & XASS_FLAG_GOURAUD)
							{
								if(p_mtd->flag & XASS_FLAG_TEXTURED)
								{
									int is_backwards = 0;
									//int is_even = p_mtd->polygt4_cnt%2 ? 0 : 1;
									//int break_idx = is_even ? p_mtd->polygt4_cnt-3 : p_mtd->polygt4_cnt-1; /* idk */
									//if(!polygt4_index || (polygt4_index == p_mtd->polygt4_cnt-1))
									//{
									//	setRGB0(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
									//	setRGB1(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									//	setRGB2(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									//	setRGB3(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									//}
									//else
									{
										setRGB0(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
										setRGB1(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
										setRGB2(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
										setRGB3(&p_mtd->apb.p_polygt4[polygt4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									}
									setUV4(&p_mtd->apb.p_polygt4[polygt4_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui3].u, uv[ui3].v,
										uv[ui2].u, uv[ui2].v);

									p_mtd->apb.p_polygt4[polygt4_index].tpage	= p_mtd->tpage;
									p_mtd->apb.p_polygt4[polygt4_index].clut	= p_mtd->clut;
									polygt4_index++;
								}
								else
								{
									/*
									 * I'm not sure if this is a problem with blender 2.76b or with my
									 * script, but the first polyg4 doesn't export the colors correctly.
									 * They're set backwards for some reason.
									 */
									if(!polyg4_index)
									{
										setRGB0(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
										setRGB1(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
										setRGB2(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
										setRGB3(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									}
									else
									{
										setRGB0(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
										setRGB1(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
										setRGB2(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
										setRGB3(&p_mtd->apb.p_polyg4[polyg4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									}
									polyg4_index++;
								}
							}
							else
							{
								if(p_mtd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_mtd->apb.p_polyft4[polyft4_index], 128, 128, 128);
									setUV4(&p_mtd->apb.p_polyft4[polyft4_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui3].u, uv[ui3].v,
										uv[ui2].u, uv[ui2].v);

									p_mtd->apb.p_polyft4[polyft4_index].tpage	= p_mtd->tpage;
									p_mtd->apb.p_polyft4[polyft4_index].clut	= p_mtd->clut;
									polyft4_index++;
								}
								else
								{
									setRGB0(&p_mtd->apb.p_polyf4[polyf4_index], r, g, b);
									polyf4_index++;
								}
							}
							break;
						default:
							printf("mtd_alloc() error: corrupted MTD file! (unknown polygon type)");
							while(1);
							break;
					}

					/* set indices for current polygon */
					p_mtd->p_indices[i].v[0] = vi0;
					p_mtd->p_indices[i].v[1] = vi1;
					p_mtd->p_indices[i].v[2] = vi2;
					p_mtd->p_indices[i].v[3] = vi3;

					if(p_mtd->flag & XASS_FLAG_LIGHTING)
					{
						p_mtd->p_indices[i].n[0] = ni0;
						p_mtd->p_indices[i].n[1] = ni1;
						p_mtd->p_indices[i].n[2] = ni2;
						p_mtd->p_indices[i].n[3] = ni3;
					}
				}
			}
		}

	}
	else
	{
		printf("mtd_alloc() error: unsupported version of mtd! (v%d)\n", p_mtd->version);
		while(1);
	}
}

/*
 * mtd_free:
 *	Frees the memory allocated by mtd_alloc.
 */
void
mtd_free(p_mtd)
struct AssetMTD *p_mtd;
{
	allocmem_free(ALLOCBUF, (char *)p_mtd->p_vertex);
	allocmem_free(ALLOCBUF, (char *)p_mtd->p_indices);
	allocmem_free(ALLOCBUF, (char *)p_mtd->p_normal);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyf3);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyf4);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyft3);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyft4);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyg3);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polyg4);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polygt3);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_polygt4);
	allocmem_free(ALLOCBUF, (char *)p_mtd->apb.p_linef2);
}

/*
 * atd_alloc:
 *	Reads an ATD file from a buffer in system memory onto
 *	a given AssetATD structure. ATD files hold 3D model
 *	data without animation.
 */
void
atd_alloc(p_atd, p_data)
struct AssetATD *p_atd;
void *p_data;
{
#define BLOCK_SIZE_KEYFRAME	16
#define BLOCK_SIZE_LINEF2	8

	int			i, j;
	unsigned char		*p_ptr = (unsigned char *)p_data;
	CVECTOR			*p_rgb;
	unsigned int		rgb_section_bnum;

	unsigned short		polyf3_index = 0;
	unsigned short		polyf4_index = 0;
	unsigned short		polyft3_index = 0;
	unsigned short		polyft4_index = 0;

	unsigned short		polyg3_index = 0;
	unsigned short		polyg4_index = 0;
	unsigned short		polygt3_index = 0;
	unsigned short		polygt4_index = 0;
	struct Vector2u		uv[USHRT_MAX];

	setVector(&p_atd->rotmod2, 0, 0, 0);

	/* init */
	p_atd->linef2_cnt = 0;
	p_atd->polyf3_cnt = 0;
	p_atd->polyf4_cnt = 0;
	p_atd->polyft3_cnt = 0;
	p_atd->polyft4_cnt = 0;
	p_atd->polyg3_cnt = 0;
	p_atd->polyg4_cnt = 0;
	p_atd->polygt3_cnt = 0;
	p_atd->polygt4_cnt = 0;

	/* atd-specific */
	p_atd->timeline_cursor = 0;
	p_atd->displayed_keyframe = 0;

	/* set header information */
	p_atd->id = *((unsigned int *)(p_ptr));
	p_ptr += 4;

	p_atd->version = *((unsigned short *)(p_ptr));
	p_ptr += 2;

	/* NOTE: supports latest version only */
	if(p_atd->version == ATD_LATEST_VERSION)
	{
		p_atd->flag = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		p_atd->nvert = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		if(p_atd->nvert > MAX_VERTS)
		{
			printf("atd_alloc() error: number of vertices exceed maximum amount (%d/%d)\n", p_atd->nvert, MAX_VERTS);
			while(1);
		}

		p_atd->npoly = *((unsigned short *)(p_ptr));
		p_ptr += 2;

		p_atd->nkey = *((unsigned char *)(p_ptr));
		p_ptr += 1;

		p_atd->fps = *((unsigned char *)(p_ptr));
		p_ptr += 1;

		p_atd->intr = *((unsigned char *)(p_ptr));
		p_ptr += 1;

		p_ptr += 1; /* skip unused */

		printf("flag:%d\nnvert:%d\nnpoly:%d\nnkey:%d\nfps:%d\nintr:%d\n", p_atd->flag, p_atd->nvert, p_atd->npoly, p_atd->nkey, p_atd->fps, p_atd->intr);

		/* init and set vertex information */
		p_atd->p_vertex = (struct ATD_VERTEX *)allocmem(ALLOCBUF, p_atd->nvert, sizeof(struct ATD_VERTEX));
		if(!p_atd->p_vertex)
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->p_vertex\n");
			printf("bytes requested: %d\n", p_atd->nvert*sizeof(VECTOR));
			while(1);
		}

		/*
		 * Set vertex information.
		 * Remember, each vertex has the same number of keyframes!
		 */
		p_atd->vbnum = *((int *)(p_ptr));
		p_ptr += 4;

		for(i = 0; i < p_atd->nvert; i++)
		{
			/*
			 * Create an array of keyframes for each vertex.
			 * The framenum and vector is separated to make
			 * the process a bit cleaner in atd_add()
			 */
			//p_atd->p_vertex[i].p_keyframe_framenum = (int *)allocmem(ALLOCBUF, p_atd->nvert, sizeof(int));
			//p_atd->p_vertex[i].p_keyframe_vector = (VECTOR *)allocmem(ALLOCBUF, p_atd->nvert, sizeof(VECTOR));

			p_atd->p_vertex[i].p_keyframe_framenum = (int *)allocmem(ALLOCBUF, p_atd->nkey, sizeof(int));
			p_atd->p_vertex[i].p_keyframe_vector = (VECTOR *)allocmem(ALLOCBUF, p_atd->nkey, sizeof(VECTOR));

			if(!p_atd->p_vertex[i].p_keyframe_framenum || !p_atd->p_vertex[i].p_keyframe_vector)
			{
				printf("atd_alloc() error: could not allocate memory for keyframe! (on vertex %d)\n", i);
				while(1);
			}

			for(j = 0; j < p_atd->nkey; j++, p_ptr += BLOCK_SIZE_KEYFRAME)
			{
				p_atd->p_vertex[i].p_keyframe_framenum[j] = *((unsigned int *)(p_ptr));

				p_atd->p_vertex[i].p_keyframe_vector[j].vx = *((unsigned int *)(p_ptr+4));
				p_atd->p_vertex[i].p_keyframe_vector[j].vy = *((unsigned int *)(p_ptr+8));
				p_atd->p_vertex[i].p_keyframe_vector[j].vz = *((unsigned int *)(p_ptr+12));
			}
		}

		/*
		 * store the UV section for later when we set the UVs
		 * of our primitives
		 */

		if(!(p_atd->flag & XASS_FLAG_WIREFRAME))
		{
			if(p_atd->flag & XASS_FLAG_TEXTURED)
			{
				unsigned int bnum = *((unsigned int *)(p_ptr));
				p_ptr += 4;

				/* FIXME: instead of storing everything into this local uv buffer, just use a fucking pointer */
				for(i = 0; i < (bnum-4)/2; i++, p_ptr += 2)
				{
					uv[i].u = *(p_ptr);
					uv[i].v = *(p_ptr+1);
				}

				if((i*2)%4) /* the iterator's value after a for loop in C89 and Python 3.4.2 (via range()) are not the same :p */
				{
					p_ptr += 2;
				}
			}
		printf("end\n");

			/*
			 * NORMAL section must be allocated in a separate memory location
			 * than our primitives since they do not store normals.
			 */
			if(p_atd->flag & XASS_FLAG_LIGHTING)
			{
				unsigned int bnum = *((unsigned int *)(p_ptr));
				unsigned int nnorm = (bnum-4)/12;

				printf("nnorm: %d; bnum: %d\n", nnorm, bnum);

				p_ptr += 4;

				p_atd->p_normal = (VECTOR *)allocmem(ALLOCBUF, nnorm, sizeof(VECTOR));
				if(!p_atd->p_normal)
				{
					printf("atd_alloc() error: could not allocate memory for p_atd->p_normal\n");
					while(1);
				}

				for(i = 0; i < nnorm; i++, p_ptr += 12)
				{
					p_atd->p_normal[i].vx = *((int *)(p_ptr));
					p_atd->p_normal[i].vy = *((int *)(p_ptr+4));
					p_atd->p_normal[i].vz = *((int *)(p_ptr+8));
				}
			}

			/* RGB section */
			if(p_atd->flag & XASS_FLAG_GOURAUD)
			{
				rgb_section_bnum = *((unsigned int *)(p_ptr));
				p_rgb = p_ptr + 4;

				p_ptr += rgb_section_bnum;
			}
		}

		/* initialize p_indices array */
		allocmem_align(ALLOCBUF);
		p_atd->p_indices = (struct AVInd *)allocmem(ALLOCBUF, p_atd->npoly, sizeof(struct AVInd));

		if(!p_atd->p_indices)
		{
			allocmem_print(ALLOCBUF);
			printf("atd_alloc() error: could not allocate memory for p_atd->p_indices (0x%.8x)\n", p_atd->p_indices);
			printf("bytes requested: %d\nallocmem_error code %d\n", p_atd->npoly*sizeof(struct AVInd), allocmem_error);
			while(1);
		}

		/*
		 * set polygon information
		 * first, count how many primitives of each type there are
		 */
		for(i = 0; i < p_atd->npoly; i++)
		{
			unsigned char	type = 0;
			unsigned int	block_size = determine_block_size(p_atd->flag);

			if(p_atd->flag & XASS_FLAG_WIREFRAME) /* use LINE_F2s, period. no mix of tris and quads */
			{
				p_atd->linef2_cnt = p_atd->npoly;
				break;
			}
			else /* normal quads and tris */
			{
				type = *( p_ptr + i*block_size );

				switch(type)
				{
					case 0: /* tri */
						if(p_atd->flag & XASS_FLAG_GOURAUD)
						{
							if(p_atd->flag & XASS_FLAG_TEXTURED)
							{
								p_atd->polygt3_cnt++;
							}
							else
							{
								p_atd->polyg3_cnt++;
							}
						}
						else
						{
							if(p_atd->flag & XASS_FLAG_TEXTURED)
							{
								p_atd->polyft3_cnt++;
							}
							else
							{
								p_atd->polyf3_cnt++;
							}
						}
						p_atd->p_indices[i].type = 0;

						break;
					case 1: /* quad */
						if(p_atd->flag & XASS_FLAG_GOURAUD)
						{
							if(p_atd->flag & XASS_FLAG_TEXTURED)
							{
								p_atd->polygt4_cnt++;
							}
							else
							{
								p_atd->polyg4_cnt++;
							}
						}
						else
						{
							if(p_atd->flag & XASS_FLAG_TEXTURED)
							{
								p_atd->polyft4_cnt++;
							}
							else
							{
								p_atd->polyf4_cnt++;
							}
						}
						p_atd->p_indices[i].type = 1;

						break;
					default:
						printf("atd_alloc() error: unknown polygon type! (%d; on iter %d @ offset 0x%x)\n", type, i, p_ptr-(unsigned char *)p_data);
						while(1);
				}
			}
		}

		/* allocate polygon primitive buffers */
		p_atd->apb.p_linef2 = (LINE_F2 *)allocmem(ALLOCBUF, p_atd->linef2_cnt, sizeof(LINE_F2));
		p_atd->apb.p_polyf3 = (POLY_F3 *)allocmem(ALLOCBUF, p_atd->polyf3_cnt, sizeof(POLY_F3));
		p_atd->apb.p_polyf4 = (POLY_F4 *)allocmem(ALLOCBUF, p_atd->polyf4_cnt, sizeof(POLY_F4));
		p_atd->apb.p_polyft3 = (POLY_FT3 *)allocmem(ALLOCBUF, p_atd->polyft3_cnt, sizeof(POLY_FT3));
		p_atd->apb.p_polyft4 = (POLY_FT4 *)allocmem(ALLOCBUF, p_atd->polyft4_cnt, sizeof(POLY_FT4));

		p_atd->apb.p_polyg3 = (POLY_G3 *)allocmem(ALLOCBUF, p_atd->polyg3_cnt, sizeof(POLY_G3));
		p_atd->apb.p_polyg4 = (POLY_G4 *)allocmem(ALLOCBUF, p_atd->polyg4_cnt, sizeof(POLY_G4));
		p_atd->apb.p_polygt3 = (POLY_GT3 *)allocmem(ALLOCBUF, p_atd->polygt3_cnt, sizeof(POLY_GT3));
		p_atd->apb.p_polygt4 = (POLY_GT4 *)allocmem(ALLOCBUF, p_atd->polygt4_cnt, sizeof(POLY_GT4));

		/* flat */
		if( (p_atd->linef2_cnt > 0) && (!p_atd->apb.p_linef2) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_linef2\n");
			while(1);
		}

		if( (p_atd->polyf3_cnt > 0) && (!p_atd->apb.p_polyf3) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyf3\n");
			while(1);
		}

		if( (p_atd->polyf4_cnt > 0) && !(p_atd->apb.p_polyf4) )
		{
			printf("bytes requsted: %d\n", p_atd->polyf4_cnt*sizeof(POLY_F4));
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyf4\n");
			while(1);
		}

		if( (p_atd->polyft3_cnt > 0) && (!p_atd->apb.p_polyft3) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyft3\n");
			while(1);
		}

		if( (p_atd->polyft4_cnt > 0) && !(p_atd->apb.p_polyft4) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyft4\n");
			while(1);
		}

		/* gouraud */
		if( (p_atd->polyg3_cnt > 0) && (!p_atd->apb.p_polyg3) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyg3\n");
			while(1);
		}

		if( (p_atd->polyg4_cnt > 0) && !(p_atd->apb.p_polyg4) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polyg4\n");
			while(1);
		}

		if( (p_atd->polygt3_cnt > 0) && (!p_atd->apb.p_polygt3) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polygt3\n");
			while(1);
		}

		if( (p_atd->polygt4_cnt > 0) && !(p_atd->apb.p_polygt4) )
		{
			printf("atd_alloc() error: could not allocate memory for p_atd->apb.p_polygt4\n");
			while(1);
		}

		/* initialize the primitive buffers */
		for(i = 0; i < p_atd->linef2_cnt; i++)
		{
			setLineF2(p_atd->apb.p_linef2 + i);
		}

		for(i = 0; i < p_atd->polyf3_cnt; i++)
		{
			setPolyF3(p_atd->apb.p_polyf3 + i);
		}

		for(i = 0; i < p_atd->polyf4_cnt; i++)
		{
			setPolyF4(p_atd->apb.p_polyf4 + i);
		}

		for(i = 0; i < p_atd->polyft3_cnt; i++)
		{
			setPolyFT3(p_atd->apb.p_polyft3 + i);
		}

		for(i = 0; i < p_atd->polyft4_cnt; i++)
		{
			setPolyFT4(p_atd->apb.p_polyft4 + i);
		}

		for(i = 0; i < p_atd->polyg3_cnt; i++)
		{
			setPolyG3(p_atd->apb.p_polyg3 + i);
		}

		for(i = 0; i < p_atd->polyg4_cnt; i++)
		{
			setPolyG4(p_atd->apb.p_polyg4 + i);
		}

		for(i = 0; i < p_atd->polygt3_cnt; i++)
		{
			setPolyGT3(p_atd->apb.p_polygt3 + i);
		}

		for(i = 0; i < p_atd->polygt4_cnt; i++)
		{
			setPolyGT4(p_atd->apb.p_polygt4 + i);
		}


		/*
		 * Now transfer data of each polygon into handler.
		 * note how we don't transfer coordinate data, as the
		 * coordinates stored in the p_atd->p_vertex array are
		 * in raw 3D coordinates. the coordinates for a POLY_XX
		 * are in screen coordinates. we only transfer RGB, UV,
		 * and index data.
		 */
		{
			int block_size;
			if(p_atd->flag & XASS_FLAG_WIREFRAME)
			{
				for(i = 0; i < p_atd->npoly; i++, p_ptr += BLOCK_SIZE_LINEF2)
				{
					unsigned char	r, g, b;
					unsigned short	vi0, vi1;

					r = *(p_ptr+1);
					g = *(p_ptr+2);
					b = *(p_ptr+3);

					vi0 = *((unsigned short *)(p_ptr+4));
					vi1 = *((unsigned short *)(p_ptr+6));

					//setRGB0(&p_atd->apb.p_linef2[i], r, g, b);
					setRGB0(&p_atd->apb.p_linef2[i], 255, 255, 255);

					p_atd->p_indices[i].v[0] = vi0;
					p_atd->p_indices[i].v[1] = vi1;
				}
			}
			else
			{
				/*
				 * Polygon sizes differ depending on certain flag
				 * settings.
				 */
				block_size = determine_block_size(p_atd->flag);

				for(i = 0; i < p_atd->npoly; i++, p_ptr += block_size)
				{
					/* FIXME: use an array :p */
					unsigned short	vi0, vi1, vi2, vi3;	/* Vertex indices */
					unsigned short	ui0, ui1, ui2, ui3;	/* UV indices */
					unsigned short	ni0, ni1, ni2, ni3;	/* normal indices */
					unsigned short	ri0, ri1, ri2, ri3;	/* rgb indices */
					unsigned char	r, g, b;

					unsigned char	*p_optsect = p_ptr+12;

					r = *(p_ptr+1);
					g = *(p_ptr+2);
					b = *(p_ptr+3);

					vi0 = *((unsigned short *)(p_ptr+4));
					vi1 = *((unsigned short *)(p_ptr+6));
					vi2 = *((unsigned short *)(p_ptr+8));
					vi3 = *((unsigned short *)(p_ptr+10));

					if(p_atd->flag & XASS_FLAG_TEXTURED)
					{
						ui0 = *((unsigned short *)(p_optsect));
						ui1 = *((unsigned short *)(p_optsect+2));
						ui2 = *((unsigned short *)(p_optsect+4));
						ui3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					if(p_atd->flag & XASS_FLAG_LIGHTING)
					{
						ni0 = *((unsigned short *)(p_optsect));
						ni1 = *((unsigned short *)(p_optsect+2));
						ni2 = *((unsigned short *)(p_optsect+4));
						ni3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					if(p_atd->flag & XASS_FLAG_GOURAUD)
					{
						ri0 = *((unsigned short *)(p_optsect));
						ri1 = *((unsigned short *)(p_optsect+2));
						ri2 = *((unsigned short *)(p_optsect+4));
						ri3 = *((unsigned short *)(p_optsect+6));
						p_optsect += 8;
					}

					/* set the colour depending on type */
					switch( *p_ptr )
					{
						case 0: /* tri */
							if(p_atd->flag & XASS_FLAG_GOURAUD)
							{
								if(p_atd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_atd->apb.p_polygt3[polygt3_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									setRGB1(&p_atd->apb.p_polygt3[polygt3_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									setRGB2(&p_atd->apb.p_polygt3[polygt3_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);

									setUV3(&p_atd->apb.p_polygt3[polygt3_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui2].u, uv[ui2].v);
									p_atd->apb.p_polygt3[polygt3_index].tpage	= p_atd->tpage;
									p_atd->apb.p_polygt3[polygt3_index].clut	= p_atd->clut;
									polygt3_index++;
								}
								else
								{
									setRGB0(&p_atd->apb.p_polyg3[polyg3_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									setRGB1(&p_atd->apb.p_polyg3[polyg3_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									setRGB2(&p_atd->apb.p_polyg3[polyg3_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									polyg3_index++;
								}
							}
							else
							{
								if(p_atd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_atd->apb.p_polyft3[polyft3_index], r, g, b);
									setUV3(&p_atd->apb.p_polyft3[polyft3_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui2].u, uv[ui2].v);
									p_atd->apb.p_polyft3[polyft3_index].tpage	= p_atd->tpage;
									p_atd->apb.p_polyft3[polyft3_index].clut	= p_atd->clut;
									polyft3_index++;
								}
								else
								{
									setRGB0(&p_atd->apb.p_polyf3[polyf3_index], r, g, b);
									polyf3_index++;
								}
							}
							break;
						case 1: /* quad */
							if(p_atd->flag & XASS_FLAG_GOURAUD)
							{
								if(p_atd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_atd->apb.p_polygt4[polygt4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									setRGB1(&p_atd->apb.p_polygt4[polygt4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
									setRGB2(&p_atd->apb.p_polygt4[polygt4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
									setRGB3(&p_atd->apb.p_polygt4[polygt4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									setUV4(&p_atd->apb.p_polygt4[polygt4_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui3].u, uv[ui3].v,
										uv[ui2].u, uv[ui2].v);

									p_atd->apb.p_polygt4[polygt4_index].tpage	= p_atd->tpage;
									p_atd->apb.p_polygt4[polygt4_index].clut	= p_atd->clut;
									polygt4_index++;
								}
								else
								{
									/*
									 * I'm not sure if this is a problem with blender 2.76b or with my
									 * script, but the first polyg4 doesn't export the colors correctly.
									 * They're set backwards for some reason.
									 */
									if(!polyg4_index)
									{
										setRGB0(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
										setRGB1(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
										setRGB2(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
										setRGB3(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
									}
									else
									{
										setRGB0(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri0].r, p_rgb[ri0].g, p_rgb[ri0].b);
										setRGB1(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri1].r, p_rgb[ri1].g, p_rgb[ri1].b);
										setRGB2(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri3].r, p_rgb[ri3].g, p_rgb[ri3].b);
										setRGB3(&p_atd->apb.p_polyg4[polyg4_index], p_rgb[ri2].r, p_rgb[ri2].g, p_rgb[ri2].b);
									}
									polyg4_index++;
								}
							}
							else
							{
								if(p_atd->flag & XASS_FLAG_TEXTURED)
								{
									setRGB0(&p_atd->apb.p_polyft4[polyft4_index], 128, 128, 128);
									setUV4(&p_atd->apb.p_polyft4[polyft4_index],
										uv[ui0].u, uv[ui0].v,
										uv[ui1].u, uv[ui1].v,
										uv[ui3].u, uv[ui3].v,
										uv[ui2].u, uv[ui2].v);

									p_atd->apb.p_polyft4[polyft4_index].tpage	= p_atd->tpage;
									p_atd->apb.p_polyft4[polyft4_index].clut	= p_atd->clut;
									polyft4_index++;
								}
								else
								{
									setRGB0(&p_atd->apb.p_polyf4[polyf4_index], r, g, b);
									polyf4_index++;
								}
							}
							break;
						default:
							printf("atd_alloc() error: corrupted atd file! (unknown polygon type)");
							while(1);
							break;
					}

					/* set indices for current polygon */
					p_atd->p_indices[i].v[0] = vi0;
					p_atd->p_indices[i].v[1] = vi1;
					p_atd->p_indices[i].v[2] = vi2;
					p_atd->p_indices[i].v[3] = vi3;

					if(p_atd->flag & XASS_FLAG_LIGHTING)
					{
						p_atd->p_indices[i].n[0] = ni0;
						p_atd->p_indices[i].n[1] = ni1;
						p_atd->p_indices[i].n[2] = ni2;
						p_atd->p_indices[i].n[3] = ni3;
					}
				}
			}
		}

	}
	else
	{
		printf("atd_alloc() error: unsupported version of atd! (v%d)\n", p_atd->version);
		while(1);
	}
}

/*
 * atd_free:
 *	Frees the memory allocated by atd_alloc.
 */
void
atd_free(p_atd)
struct AssetATD *p_atd;
{
	int i, j;

	allocmem_free(ALLOCBUF, (char *)p_atd->p_vertex);
	allocmem_free(ALLOCBUF, (char *)p_atd->p_indices);
	allocmem_free(ALLOCBUF, (char *)p_atd->apb.p_polyf3);
	allocmem_free(ALLOCBUF, (char *)p_atd->apb.p_polyf4);
	allocmem_free(ALLOCBUF, (char *)p_atd->apb.p_polyft3);
	allocmem_free(ALLOCBUF, (char *)p_atd->apb.p_polyft4);
	allocmem_free(ALLOCBUF, (char *)p_atd->apb.p_linef2);

	/* ATD-specific */
	for(i = 0; i < p_atd->nvert; i++)
	{
		allocmem_free(ALLOCBUF, (char *)p_atd->p_vertex[i].p_keyframe_framenum);
		allocmem_free(ALLOCBUF, (char *)p_atd->p_vertex[i].p_keyframe_vector);
	}
	allocmem_free(ALLOCBUF, (char *)p_atd->p_vertex);
}

/*
 * atd_rgb:
 *	Sets the color of every* polygon in an ATD.
 *
 *	*not really
 */
atd_rgb(p_atd, r, g, b)
struct AssetATD *p_atd;
unsigned char r, g, b;
{
	int i;

	for(i = 0; i < p_atd->linef2_cnt; i++)
	{
		setRGB0(&p_atd->apb.p_linef2[i], r, g, b);
	}
}

/*
 * atd_play:
 *	Plays a given ATD file.
 */
atd_play(p_atd, is_loop)
struct AssetATD *p_atd;
int is_loop;
{
	int i;

	int max_framenum = p_atd->p_vertex[0].p_keyframe_framenum[p_atd->nkey-1]; /* FIXME: does not need to be set here every frame!!!! */
	p_atd->timeline_cursor++;

	/* set timeline_cursor to max if it is greater than max */
	if(p_atd->timeline_cursor >= max_framenum)
	{
		if( is_loop )
		{
			p_atd->timeline_cursor = 0;
		}
		else
		{
			p_atd->timeline_cursor = max_framenum;
		}
	}

	/* ATD FIXME: have a TIMELINE section */
	for(i = 0; i < p_atd->nkey; i++)
	{
		if(p_atd->p_vertex[0].p_keyframe_framenum[i] <= p_atd->timeline_cursor)
		{
			p_atd->displayed_keyframe = i;
		}
		else
		{
			break;
		}
	}
}

/*
 * set_aabb_assquad:
 *	Helper function for cxc_alloc() similar to
 *	set_aabb_cuboid(), but instead only finds aabb
 *	in 2d (x and z axes), and only cycles through
 *	4 vertices instead of 24.
 */
static void
set_aabb_assquad(p_assquad)
struct AssQuad *p_assquad;
{
	int		i;
	short		vx_lo, vy_lo, vz_lo;
	short		vx_hi, vy_hi, vz_hi;

	for(i = 0; i < 2; i++)
	{
		p_assquad->aabb_x[i] = 0;
		p_assquad->aabb_y[i] = 0;
		p_assquad->aabb_z[i] = 0;
	}

	/* determine axes by cycling through all 4 vertices */
	/* FIXME: probably a nicer way to do this... */
	for(i = 0; i < 4; i++)
	{
		if(i == 0)
		{
			vx_lo = vx_hi = p_assquad->global_coords.v[i].vx;
			vy_lo = vy_hi = p_assquad->global_coords.v[i].vy;
			vz_lo = vz_hi = p_assquad->global_coords.v[i].vz;
			continue;
		}
		else
		{
			short vx = p_assquad->global_coords.v[i].vx;
			short vy = p_assquad->global_coords.v[i].vy;
			short vz = p_assquad->global_coords.v[i].vz;

			/*** x axis ***/
			if(vx > vx_hi) /* left */
			{
				vx_hi = vx;
			}
			if(vx < vx_lo) /* right */
			{
				vx_lo = vx;
			}

			/*** y axis ***/
			if(vy > vy_lo) /* bottom */
			{
				vy_lo = vy;
			}
			if(vy < vy_hi) /* top */
			{
				vy_hi = vy;
			}

			/*** z axis ***/
			if(vz < vz_lo) /* front */
			{
				vz_lo = vz;
			}
			if(vz > vz_hi) /* back */
			{
				vz_hi = vz;
			}
		}
	}

	p_assquad->aabb_x[0] = vx_hi;
	p_assquad->aabb_x[1] = vx_lo;

	p_assquad->aabb_y[0] = vy_lo;
	p_assquad->aabb_y[1] = vy_hi;

	p_assquad->aabb_z[0] = vz_lo;
	p_assquad->aabb_z[1] = vz_hi;
}

/*
 * set_aabb_cuboid:
 *	Helper function for cxc_alloc() that finds the
 *	aligned axes that a cuboid is bounded to. This
 *	is done via the quad's global coordinates.
 *
 *	FIXME: should be rename p_cuboid to p_asscuboid
 */
static void
set_aabb_cuboid(p_cuboid)
struct AssCuboid *p_cuboid;
{
	int		i, j;

	/* 0: lo, 1: hi */
	short		vx[2];
	short		vy[2];
	short		vz[2];

	for(i = 0; i < 2; i++)
	{
		p_cuboid->aabb_x[i] = 0;
		p_cuboid->aabb_y[i] = 0;
		p_cuboid->aabb_z[i] = 0;
	}

	/* read */
	for(i = 0; i < 6; i++)
	{
		short *p_aabb_x = p_cuboid->quad[i].aabb_x;
		short *p_aabb_y = p_cuboid->quad[i].aabb_y;
		short *p_aabb_z = p_cuboid->quad[i].aabb_z;

		set_aabb_assquad(&p_cuboid->quad[i]);

		if(i == 0)
		{
			vx[0] = p_cuboid->quad[i].aabb_x[0];
			vy[0] = p_cuboid->quad[i].aabb_y[0];
			vz[0] = p_cuboid->quad[i].aabb_z[0];

			vx[1] = p_cuboid->quad[i].aabb_x[1];
			vy[1] = p_cuboid->quad[i].aabb_y[1];
			vz[1] = p_cuboid->quad[i].aabb_z[1];
			continue;
		}

		/* x axis */
		if(p_aabb_x[0] < vx[0])
		{
			vx[0] = p_aabb_x[0];
		}
		if(p_aabb_x[1] > vx[1])
		{
			vx[1] = p_aabb_x[1];
		}

		/* y axis */
		if(p_aabb_y[0] > vy[0])
		{
			vy[0] = p_aabb_y[0];
		}
		if(p_aabb_y[1] < vy[1])
		{
			vy[1] = p_aabb_y[1];
		}

		/* z axis */
		if(p_aabb_z[0] < vz[0])
		{
			vz[0] = p_aabb_z[0];
		}
		if(p_aabb_z[1] > vz[1])
		{
			vz[1] = p_aabb_z[1];
		}
	}

	for(i = 0; i < 2; i++)
	{
		p_cuboid->aabb_x[i] = vx[i];
		p_cuboid->aabb_y[i] = vy[i];
		p_cuboid->aabb_z[i] = vz[i];
	}
}

/*
 * cxc_alloc:
 *	Reads a CXC file from a buffer in system memory onto
 *	a given AssetCXC structure. CXC files hold collision
 *	data.
 */
void
cxc_alloc(p_cxc, p_data, p_parent, parent)
struct AssetCXC *p_cxc;
void *p_data, *p_parent;
int parent;
{
	int i, j, k;

	unsigned char *p_ptr = (unsigned char *)p_data;

	/* set metadata */
	p_cxc->p_parent = p_parent;
	p_cxc->parent = parent;

	/* read HEADER */
	p_cxc->id = *((unsigned int *)(p_ptr));
	p_ptr += 4;

	/* skip reserved */
	p_ptr += 2;

	p_cxc->version = *((unsigned short *)(p_ptr));
	p_ptr += 2;

	/* NOTE: supports latest version only */
	if(p_cxc->version == CXC_LATEST_VERSION)
	{
		/* read & allocate COLMAP */
		p_cxc->COLMAP.bnum = *((unsigned int *)(p_ptr));
		p_cxc->COLMAP.nquads = *((unsigned short *)(p_ptr+4));
		p_cxc->COLMAP.ncuboids = *((unsigned short *)(p_ptr+6));
		p_ptr += 8;

		p_cxc->COLMAP.p_quad = (struct AssQuad *)allocmem(ALLOCBUF, p_cxc->COLMAP.nquads, sizeof(struct AssQuad));
		p_cxc->COLMAP.p_cuboid = (struct AssCuboid *)allocmem(ALLOCBUF, p_cxc->COLMAP.ncuboids, sizeof(struct AssCuboid));

		if(p_cxc->COLMAP.nquads > 0 && !p_cxc->COLMAP.p_quad)
		{
			printf("cxc_alloc() error: could not allocate memory for p_cxc->COLMAP.p_quad\n");
			while(1);
		}

		if(p_cxc->COLMAP.ncuboids > 0 && !p_cxc->COLMAP.p_cuboid)
		{
			printf("bytes requested: %d\n", p_cxc->COLMAP.ncuboids*sizeof(struct AssCuboid));
			printf("cxc_alloc() error: could not allocate memory for p_cxc->COLMAP.p_cuboid\n");
			while(1);
		}

		/* FIXME: style inconsistencies with pointer names */
		/* read COLMAP section's QUAD & CUBOID data */
		{
			struct AssQuad *q;
			struct AssCuboid *c;
			SVECTOR v[4];

			/* QUADs */
			for(i = 0; i < p_cxc->COLMAP.nquads; i++)
			{
				q = &p_cxc->COLMAP.p_quad[i];

				setVector(&q->translation, 0, 0, TRANS_Z);

				for(j = 0; j < 4; j++)
				{
					setLineF2(&q->prim[j]);
					setRGB0(&q->prim[j], 255, 128, 128);
					q->global_coords.v[j].vx = *((unsigned short *)(p_ptr));
					q->global_coords.v[j].vy = *((unsigned short *)(p_ptr+2));
					q->global_coords.v[j].vz = *((unsigned short *)(p_ptr+4));

					v[j] = q->global_coords.v[j];

					p_ptr += 6;
				}

				/* determine slope direction via those four vertices */
				{
					int delta = 0; /* highest vertical delta */
					int tmp;

					for(j = 1; j < 4; j++)
					{
						if( (tmp = ABS(v[0].vy - v[j].vy)) > delta )
						{
							delta = tmp;

							/*
							 * This is a rather primitive slope collision setup, but
							 * it'll do for now. Collision that does not entirely rely
							 * upon one axis is not possible with this solution.
							 */
							if( ABS(v[0].vx - v[j].vx) > ABS(v[0].vz - v[j].vz) )
							{
								q->global_coords.direction = CXC_SLOPE_X;
								q->global_coords.is_flipped = sign(v[0].vx - v[j].vx);
							}
							else
							{
								q->global_coords.direction = CXC_SLOPE_Z;
								q->global_coords.is_flipped = sign(v[0].vz - v[j].vz);
							}

							break;
						}
						else
						{
							q->global_coords.direction = CXC_SLOPE_FLAT;
						}
					}
				}

				/* set 2d aabb's */
				set_aabb_assquad(q);

				/* initialize debug prims for aabb */
				{
					int r = (rand()*32)>>12;
					int g = (rand()*32)>>12;
					int b = (rand()*32)>>12;

					for(j = 0; j < 2; j++)
					{
						setTile8(&p_cxc->COLMAP.p_quad[i].aabb_corner[j]);
						setRGB0(&p_cxc->COLMAP.p_quad[i].aabb_corner[j], r, g, b);
					}
				}
			}

			/* CUBOIDs */
			for(i = 0; i < p_cxc->COLMAP.ncuboids; i++)
			{
				c = &p_cxc->COLMAP.p_cuboid[i];

				/* 6 QUADs per CUBOID */
				for(j = 0; j < 6; j++)
				{
					setVector(&c->quad[j].translation, 0, 0, TRANS_Z);

					for(k = 0; k < 4; k++)
					{
						setLineF2(&c->quad[j].prim[k]);
						setRGB0(&c->quad[j].prim[k], 255, 128, 128);
						c->quad[j].global_coords.v[k].vx = *((short *)(p_ptr));
						c->quad[j].global_coords.v[k].vy = *((short *)(p_ptr+2));
						c->quad[j].global_coords.v[k].vz = *((short *)(p_ptr+4));

						p_ptr += 6;
					}
				}

				set_aabb_cuboid(c);
			}
		}

		/* read & allocate AREATRIG */
		p_cxc->AREATRIG.bnum = *((unsigned int *)(p_ptr));
		p_cxc->AREATRIG.ncuboids = *((unsigned short *)(p_ptr+6));
		p_ptr += 8;

		p_cxc->AREATRIG.p_cuboid = (struct AssCuboid *)allocmem(ALLOCBUF, p_cxc->AREATRIG.ncuboids, sizeof(struct AssCuboid));

		if(p_cxc->AREATRIG.ncuboids > 0 && !p_cxc->AREATRIG.p_cuboid)
		{
			allocmem_print(ALLOCBUF);
			printf("cxc_alloc() error: could not allocate memory for p_cxc->AREATRIG.p_cuboid\n");
			printf("bytes requested: %d\n", p_cxc->AREATRIG.ncuboids*sizeof(struct AssCuboid));
			while(1);
		}

		/* read AREATRIG section's CUBOID data */
		{
			struct AssCuboid *c;

			for(i = 0; i < p_cxc->AREATRIG.ncuboids; i++)
			{
				c = &p_cxc->AREATRIG.p_cuboid[i];
				for(j = 0; j < 6; j++)
				{
					setVector(&c->quad[j].translation, 0, 0, TRANS_Z);

					for(k = 0; k < 4; k++)
					{
						setLineF2(&c->quad[j].prim[k]);
						setRGB0(&c->quad[j].prim[k], 128, 128, 255);
						c->quad[j].global_coords.v[k].vx = *((short *)(p_ptr));
						c->quad[j].global_coords.v[k].vy = *((short *)(p_ptr+2));
						c->quad[j].global_coords.v[k].vz = *((short *)(p_ptr+4));

						p_ptr += 6;
					}
				}

				set_aabb_cuboid(c);
			}
		}
	}
}

/*
 * cxc_free:
 *	Frees the memory allocated by cxc_alloc.
 */
cxc_free(p_cxc)
struct AssetCXC *p_cxc;
{
	allocmem_free(ALLOCBUF, p_cxc->COLMAP.p_quad);
	allocmem_free(ALLOCBUF, p_cxc->COLMAP.p_cuboid);
	allocmem_free(ALLOCBUF, p_cxc->AREATRIG.p_cuboid);
}

/*
 * check_pmode:
 *	Checks whether or not the pmode of a TIM file is a
 *	valid. Completely halts the program if there's a
 *	discrepancy.
 */
check_pmode(func_name, pmode)
const char *func_name;
char pmode;
{
	switch(pmode)
	{
		case 0:
		case 1:
		case 2:
			return;
		case 3:
			printf("%s() error: 24-bit textures are not supported!\n", func_name);
			break;
		case 4:
			printf("%s() error: textures with mixed CLUTs are not supported!\n", func_name);
			break;
		default:
			printf("%s() error: unknown pmode value (%d)\n", func_name, pmode);
			break;
	}
	while(1);
}

/*
 * tex_upload:
 *	Uploads all TIM images from a TEX file into VRAM.
 *	Each TIM should be 256x256px.
 */
tex_upload(p_data)
void *p_data;
{
	int i, j;
	unsigned int timtotal = *((unsigned int *)(p_data));
	char *p_tim = p_data;

	p_tim += 4+timtotal*4;

	for(i = 0; i < timtotal; i++)
	{
		char flag = *(p_tim+4);
		char pmode = flag&0x07;

		p_tim += 8;

		check_pmode("tex_upload", pmode);

		{
			unsigned int nblock = 1;
			unsigned int bnum = 0;
			RECT rect;

			if(flag & (1<<3))
			{
				nblock++;
			}

			for(j = 0; j < nblock; j++)
			{
				bnum = *((unsigned int *)(p_tim));
				setRECT( &rect, *((short *)(p_tim+4)),
					*((short *)(p_tim+6)),
					*((short *)(p_tim+8)),
					*((short *)(p_tim+10)) );
				LoadImage(&rect, (unsigned long *)(p_tim+12));
				p_tim += bnum;
			}
		}
	}
}
